# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.services.trunk.drivers import base
from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.services.trunk import constants as trunk_consts

from vmware_nsx._i18n import _
from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.common import utils as nsx_utils
from vmware_nsx.extensions import projectpluginmap
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3.policy import constants as p_constants
from vmware_nsxlib.v3 import utils as nsxlib_utils

LOG = logging.getLogger(__name__)

SUPPORTED_INTERFACES = (
    portbindings.VIF_TYPE_OVS,
)
SUPPORTED_SEGMENTATION_TYPES = (
    trunk_consts.SEGMENTATION_TYPE_VLAN,
)

DRIVER_NAME = 'vmware_nsxp_trunk'
TRUNK_ID_TAG_NAME = 'os-neutron-trunk-id'


class NsxpTrunkHandler(object):
    def __init__(self, plugin_driver):
        self.plugin_driver = plugin_driver

    def _get_port_tags_and_network(self, context, port_id):
        _, tags, net = self._get_port_compute_tags_and_net(context, port_id)
        return tags, net

    def _get_port_compute_tags_and_net(self, context, port_id):
        port = self.plugin_driver.get_port(context, port_id)
        segment_id = self.plugin_driver._get_network_nsx_segment_id(
            context, port['network_id'])
        lport = self.plugin_driver.nsxpolicy.segment_port.get(
            segment_id, port_id)
        is_compute = port.get('device_owner', '').startswith(
            constants.DEVICE_OWNER_COMPUTE_PREFIX)
        return is_compute, segment_id, lport.get('tags', [])

    def _update_tags(self, port_id, tags, tags_update, is_delete=False):
        if is_delete:
            tags = [tag for tag in tags if tag not in tags_update]
        else:
            for tag in tags:
                for tag_u in tags_update:
                    if tag_u['scope'] == tag['scope']:
                        tag['tag'] = tag_u['tag']
                        tags_update.remove(tag_u)
                        break

            tags.extend(
                [tag for tag in tags_update if tag not in tags])
            if len(tags) > nsxlib_utils.MAX_TAGS:
                LOG.warning("Cannot add external tags to port %s: "
                            "too many tags", port_id)
        return tags

    def _set_subports(self, context, parent_port_id, subports):
        for subport in subports:
            # Update port with parent port for backend.

            # Set properties for VLAN trunking
            if subport.segmentation_type == nsx_utils.NsxV3NetworkTypes.VLAN:
                seg_id = subport.segmentation_id
            else:
                msg = (_("Cannot create a subport %s with no segmentation"
                         " id") % subport.port_id)
                LOG.error(msg)
                raise nsx_exc.NsxPluginException(err_msg=msg)

            tags_update = [{'scope': TRUNK_ID_TAG_NAME,
                            'tag': subport.trunk_id}]

            segment_id, tags = self._get_port_tags_and_network(
                context, subport.port_id)

            tags = self._update_tags(
                subport.port_id, tags, tags_update, is_delete=False)

            # Update logical port in the backend to set/unset parent port
            try:
                self.plugin_driver.nsxpolicy.segment_port.attach(
                    segment_id,
                    subport.port_id,
                    p_constants.ATTACHMENT_CHILD,
                    subport.port_id,
                    context_id=parent_port_id,
                    traffic_tag=seg_id,
                    tags=tags)

            except nsxlib_exc.ManagerError as e:
                with excutils.save_and_reraise_exception():
                    LOG.error("Unable to update subport for attachment "
                              "type. Exception is %s", e)

    def _unset_subports(self, context, subports):
        for subport in subports:
            # Update port and remove parent port attachment in the backend
            # Unset the parent port properties from child port

            tags_update = [{'scope': TRUNK_ID_TAG_NAME,
                            'tag': subport.trunk_id}]

            is_compute, segment_id, tags = self._get_port_compute_tags_and_net(
                context, subport.port_id)

            tags = self._update_tags(
                subport.port_id, tags, tags_update, is_delete=True)

            # Update logical port in the backend to set/unset parent port
            vif_id = None
            if is_compute:
                vif_id = subport.port_id
            try:
                self.plugin_driver.nsxpolicy.segment_port.detach(
                    segment_id, subport.port_id, vif_id=vif_id, tags=tags)

            except nsxlib_exc.ManagerError as e:
                with excutils.save_and_reraise_exception():
                    LOG.error("Unable to update subport for attachment "
                              "type. Exception is %s", e)

    def trunk_created(self, context, trunk):
        tags_update = [{'scope': TRUNK_ID_TAG_NAME, 'tag': trunk.id}]
        segment_id, tags = self._get_port_tags_and_network(
            context, trunk.port_id)

        tags = self._update_tags(
            trunk.port_id, tags, tags_update, is_delete=False)

        try:
            self.plugin_driver.nsxpolicy.segment_port.attach(
                    segment_id,
                    trunk.port_id,
                    vif_id=trunk.port_id,
                    attachment_type=p_constants.ATTACHMENT_PARENT,
                    tags=tags)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error("Parent port attachment for trunk %(trunk)s failed "
                          "with error %(e)s", {'trunk': trunk.id, 'e': e})

        if trunk.sub_ports:
            self.subports_added(context, trunk, trunk.sub_ports)

    def trunk_deleted(self, context, trunk):
        tags_update = [{'scope': TRUNK_ID_TAG_NAME, 'tag': trunk.id}]

        is_compute, segment_id, tags = self._get_port_compute_tags_and_net(
            context, trunk.port_id)

        tags = self._update_tags(
            trunk.port_id, tags, tags_update, is_delete=True)

        try:
            vif_id = None
            if is_compute:
                vif_id = trunk.port_id
            self.plugin_driver.nsxpolicy.segment_port.detach(
                segment_id, trunk.port_id, vif_id=vif_id, tags=tags)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error("Parent port detachment for trunk %(trunk)s failed "
                          "with error %(e)s", {'trunk': trunk.id, 'e': e})

        self.subports_deleted(context, trunk, trunk.sub_ports)

    def subports_added(self, context, trunk, subports):
        try:
            self._set_subports(context, trunk.port_id, subports)
            trunk.update(status=trunk_consts.TRUNK_ACTIVE_STATUS)
        except (nsxlib_exc.ManagerError, nsxlib_exc.ResourceNotFound):
            trunk.update(status=trunk_consts.TRUNK_ERROR_STATUS)

    def subports_deleted(self, context, trunk, subports):
        try:
            self._unset_subports(context, subports)
        except (nsxlib_exc.ManagerError, nsxlib_exc.ResourceNotFound):
            trunk.update(status=trunk_consts.TRUNK_ERROR_STATUS)

    def trunk_event(self, resource, event, trunk_plugin, payload):
        if event == events.AFTER_CREATE:
            self.trunk_created(payload.context, payload.current_trunk)
        elif event == events.AFTER_DELETE:
            self.trunk_deleted(payload.context, payload.original_trunk)

    def subport_event(self, resource, event, trunk_plugin, payload):
        if event == events.AFTER_CREATE:
            self.subports_added(
                payload.context, payload.original_trunk, payload.subports)
        elif event == events.AFTER_DELETE:
            self.subports_deleted(
                payload.context, payload.original_trunk, payload.subports)


class NsxpTrunkDriver(base.DriverBase):
    """Driver to implement neutron's trunk extensions."""

    @property
    def is_loaded(self):
        try:
            plugin_type = self.plugin_driver.plugin_type()
            return plugin_type == projectpluginmap.NsxPlugins.NSX_P
        except cfg.NoSuchOptError:
            return False

    @classmethod
    def create(cls, plugin_driver):
        cls.plugin_driver = plugin_driver
        return cls(DRIVER_NAME, SUPPORTED_INTERFACES,
                   SUPPORTED_SEGMENTATION_TYPES,
                   agent_type=None, can_trunk_bound_port=True)

    @registry.receives(resources.TRUNK_PLUGIN, [events.AFTER_INIT])
    def register(self, resource, event, trigger, payload=None):
        super(NsxpTrunkDriver, self).register(
            resource, event, trigger, payload=payload)
        self._handler = NsxpTrunkHandler(self.plugin_driver)
        for event in (events.AFTER_CREATE, events.AFTER_DELETE):
            registry.subscribe(self._handler.trunk_event,
                               resources.TRUNK,
                               event)
            registry.subscribe(self._handler.subport_event,
                               resources.SUBPORTS,
                               event)
        LOG.debug("VMware NSXP trunk driver initialized.")



# Copyright 2019 VMware, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""lbaas_error_no_member

Revision ID: 99bfcb6003c6
Revises: fc6308289aca
Create Date: 2019-03-07 11:27:00.000000
"""

from alembic import op

from neutron.db import migration

# revision identifiers, used by Alembic.
revision = '99bfcb6003c6'
down_revision = 'fc6308289aca'


def upgrade():
    if (migration.schema_has_table('nsxv3_lbaas_loadbalancers') and
        migration.schema_has_table('lbaas_loadbalancers') and
        migration.schema_has_table('nsxv_lbaas_loadbalancer_bindings')):
        # Mark as ERROR loadbalancers without nsx mapping
        op.execute("UPDATE lbaas_loadbalancers "
                   "SET provisioning_status='ERROR' "
                   "WHERE id not in (SELECT loadbalancer_id FROM "
                   "nsxv3_lbaas_loadbalancers) and id not in (SELECT "
                   "loadbalancer_id FROM nsxv_lbaas_loadbalancer_bindings)")



# Copyright 2017 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import context as n_context
from neutron_lib.exceptions import firewall_v2 as exceptions
from oslo_log import log as logging

from vmware_nsx.services.fwaas.common import fwaas_driver_base

LOG = logging.getLogger(__name__)


class CommonEdgeFwaasV3Driver(fwaas_driver_base.EdgeFwaasDriverBaseV2):
    """Base class for NSX-V3/Policy driver for Firewall As A Service  V2."""

    def __init__(self, driver_name):
        super(CommonEdgeFwaasV3Driver, self).__init__(driver_name)
        self.backend_support = True
        self.driver_exception = exceptions.FirewallInternalDriverError
        self._core_plugin = None

    @property
    def core_plugin(self):
        """Get the core plugin - should be implemented by each driver"""
        pass

    def validate_backend_version(self):
        """Validate NSX backend supports FWaaS
        Can be implemented by each driver
        """
        pass

    def _update_backend_routers(self, apply_list, fwg_id):
        """Update all the affected router on the backend"""
        self.validate_backend_version()
        LOG.info("Updating routers firewall for firewall group %s", fwg_id)
        context = n_context.get_admin_context()
        routers = set()
        # the apply_list is a list of tuples: routerInfo, port-id
        for router_info, port_id in apply_list:
            # Skip dummy entries that were added only to avoid errors
            if isinstance(router_info, str):
                continue
            # Skip unsupported routers
            if not self.should_apply_firewall_to_router(router_info.router):
                continue
            routers.add(router_info.router_id)

        # update each router once
        for router_id in routers:
            try:
                self.core_plugin.update_router_firewall(context, router_id,
                                                        from_fw=True)
            except Exception as e:
                # The core plugin failed to update the firewall
                LOG.error("Failed to update NSX edge firewall for router %s: "
                          "%s", router_id, e)
                raise self.driver_exception(driver=self.driver_name)

    def should_apply_firewall_to_router(self, router_data):
        """Return True if the firewall rules should be added the router"""
        if not router_data.get('external_gateway_info'):
            LOG.info("Cannot apply firewall to router %s with no gateway",
                     router_data['id'])
            return False
        return True



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_common
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgePoolManagerFromDict(base_mgr.NsxpLoadbalancerBaseManager):
    def _get_pool_kwargs(self, name=None, tags=None, algorithm=None,
                         description=None):
        kwargs = {
            'snat_translation': {'type': "LBSnatAutoMap"}}
        if name:
            kwargs['name'] = name
        if tags:
            kwargs['tags'] = tags
        if algorithm:
            kwargs['algorithm'] = algorithm
        if description:
            kwargs['description'] = description
        return kwargs

    def _get_pool_tags(self, context, pool):
        return lb_utils.get_tags(self.core_plugin, pool['id'],
                                 lb_const.LB_POOL_TYPE, pool['tenant_id'],
                                 context.project_name)

    def _remove_persistence(self, vs_data):
        lb_utils.delete_persistence_profile(
            self.core_plugin.nsxpolicy,
            vs_data.get('lb_persistence_profile_path', ''))

    def _process_vs_update(self, context, pool, pool_id, switch_type,
                           listener, completor):
        vs_client = self.core_plugin.nsxpolicy.load_balancer.virtual_server
        try:
            # Process pool persistence profile and
            # create/update/delete profile for virtual server
            vs_data = vs_client.get(listener['id'])
            if pool and pool_id:
                (persistence_profile_id,
                 post_process_func) = lb_utils.setup_session_persistence(
                    self.core_plugin.nsxpolicy, pool,
                    self._get_pool_tags(context, pool),
                    switch_type, listener, vs_data)
            else:
                post_process_func = functools.partial(
                    self._remove_persistence, vs_data)
                persistence_profile_id = None
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error("Failed to configure session persistence "
                          "profile for pool %(pool_id)s",
                          {'pool_id': pool['id']})
        try:
            # Update persistence profile and pool on virtual server
            vs_client.update(
                listener['id'],
                pool_id=pool_id,
                lb_persistence_profile_id=persistence_profile_id)

            LOG.debug("Updated NSX virtual server %(vs_id)s with "
                      "pool %(pool_id)s and persistence profile %(prof)s",
                      {'vs_id': listener['id'], 'pool_id': pool['id'],
                       'prof': persistence_profile_id})
            if post_process_func:
                post_process_func()
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to attach pool %s to virtual '
                          'server %s', pool['id'], listener['id'])

    def create(self, context, pool, completor):
        pool_client = self.core_plugin.nsxpolicy.load_balancer.lb_pool

        pool_name = utils.get_name_and_uuid(pool['name'] or 'pool', pool['id'])
        tags = self._get_pool_tags(context, pool)

        description = pool.get('description')
        lb_algorithm = lb_const.LB_POOL_ALGORITHM_MAP.get(pool['lb_algorithm'])

        if pool.get('listeners') and len(pool['listeners']) > 1:
            completor(success=False)
            msg = (_('Failed to create pool: Multiple listeners are not '
                     'supported.'))
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        # NOTE(salv-orlando): Guard against accidental compat breakages
        try:
            listener = pool['listener'] or pool['listeners'][0]
        except IndexError:
            # If listeners is an empty list we hit this exception
            listener = None
        # Perform additional validation for session persistence before
        # creating resources in the backend
        lb_common.validate_session_persistence(pool, listener, completor)
        try:
            kwargs = self._get_pool_kwargs(
                pool_name, tags, lb_algorithm, description)
            pool_client.create_or_overwrite(lb_pool_id=pool['id'], **kwargs)
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = (_('Failed to create pool on NSX backend: %(pool)s') %
                   {'pool': pool['id']})
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        # The pool object can be created with either --listener or
        # --loadbalancer option. If listener is present, the virtual server
        # will be updated with the pool. Otherwise, just return. The binding
        # will be added later when the pool is associated with layer7 rule.
        # FIXME(salv-orlando): This two-step process can leave a zombie pool on
        # NSX if the VS update operation fails
        if listener:
            self._process_vs_update(context, pool, pool['id'], False,
                                    listener, completor)
        completor(success=True)

    def update(self, context, old_pool, new_pool, completor):
        pool_client = self.core_plugin.nsxpolicy.load_balancer.lb_pool

        pool_name = None
        tags = None
        lb_algorithm = None
        description = None
        if new_pool.get('listeners') and len(new_pool['listeners']) > 1:
            completor(success=False)
            msg = (_('Failed to update pool %s: Multiple listeners are not '
                     'supported.') % new_pool['id'])
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        if new_pool['name'] != old_pool['name']:
            pool_name = utils.get_name_and_uuid(new_pool['name'] or 'pool',
                                                new_pool['id'])
            tags = self._get_pool_tags(context, new_pool)
        if new_pool['lb_algorithm'] != old_pool['lb_algorithm']:
            lb_algorithm = lb_const.LB_POOL_ALGORITHM_MAP.get(
                new_pool['lb_algorithm'])
        if new_pool.get('description') != old_pool.get('description'):
            description = new_pool['description']
        # NOTE(salv-orlando): Guard against accidental compat breakages
        try:
            listener = new_pool['listener'] or new_pool['listeners'][0]
        except IndexError:
            # If listeners is an empty list we hit this exception
            listener = None
            # Perform additional validation for session persistence before
            # operating on resources in the backend
        lb_common.validate_session_persistence(new_pool, listener, completor)
        try:
            kwargs = self._get_pool_kwargs(pool_name, tags, lb_algorithm,
                                           description)
            pool_client.update(lb_pool_id=new_pool['id'], **kwargs)
            if (listener and new_pool['session_persistence'] !=
                old_pool['session_persistence']):
                switch_type = lb_common.session_persistence_type_changed(
                    new_pool, old_pool)
                self._process_vs_update(context, new_pool, new_pool['id'],
                                        switch_type, listener, completor)
            completor(success=True)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to update pool %(pool)s with '
                          'error %(error)s',
                          {'pool': old_pool['id'], 'error': e})

    def delete(self, context, pool, completor):
        pool_client = self.core_plugin.nsxpolicy.load_balancer.lb_pool

        # NOTE(salv-orlando): Guard against accidental compat breakages
        try:
            listener = pool.get('listener') or pool.get('listeners', [])[0]
        except IndexError:
            # If listeners is an empty list we hit this exception
            listener = None
        if listener:
            try:
                self._process_vs_update(
                    context, pool, None, False, listener, completor)
            except Exception as e:
                LOG.error('Disassociation of listener %(lsn)s from pool '
                          '%(pool)s failed with error %(err)s',
                          {'lsn': listener['id'],
                           'pool': pool['id'],
                           'err': e})
        try:
            pool_client.delete(pool['id'])
        except nsxlib_exc.ResourceNotFound:
            pass
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = (_('Failed to delete lb pool from nsx: %(pool)s') %
                   {'pool': pool['id']})
            raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        # Delete the attached health monitor as well
        if pool.get('healthmonitor'):
            hm = pool['healthmonitor']
            monitor_client = lb_utils.get_monitor_policy_client(
                self.core_plugin.nsxpolicy.load_balancer, hm)

            try:
                monitor_client.delete(hm['id'])
            except nsxlib_exc.ResourceNotFound:
                pass
            except nsxlib_exc.ManagerError as exc:
                completor(success=False)
                msg = _('Failed to delete monitor %(monitor)s from backend '
                        'with exception %(exc)s') % {'monitor': hm['id'],
                                                     'exc': exc}
                raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

        completor(success=True)

    def delete_cascade(self, context, pool, completor):
        self.delete(context, pool, completor)



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron_lib.api.definitions import constants as fwaas_consts

from vmware_nsxlib.v3 import nsx_constants

LOG = logging.getLogger(__name__)


def translate_fw_rule_action(fwaas_action, fwaas_rule_id):
    """Translate FWaaS action to NSX action"""
    if fwaas_action == fwaas_consts.FWAAS_ALLOW:
        return nsx_constants.FW_ACTION_ALLOW
    if fwaas_action == fwaas_consts.FWAAS_DENY:
        return nsx_constants.FW_ACTION_DROP
    if fwaas_action == fwaas_consts.FWAAS_REJECT:
        # reject is not supported by the NSX edge firewall
        LOG.warning("Reject action is not supported by the NSX backend "
                    "for edge firewall. Using %(action)s instead for "
                    "rule %(id)s",
              {'action': nsx_constants.FW_ACTION_DROP,
               'id': fwaas_rule_id})
        return nsx_constants.FW_ACTION_DROP
    # Unexpected action
    LOG.error("Unsupported FWAAS action %(action)s for rule %(id)s", {
        'action': fwaas_action, 'id': fwaas_rule_id})


def translate_fw_rule_protocol(fwaas_protocol):
    """Translate FWaaS L4 protocol to NSX protocol"""
    if fwaas_protocol.lower() == 'tcp':
        return nsx_constants.TCP
    if fwaas_protocol.lower() == 'udp':
        return nsx_constants.UDP
    if fwaas_protocol.lower() == 'icmp':
        # This will cover icmpv6 too, when adding  the rule.
        return nsx_constants.ICMPV4


def translate_fw_rule_ports(ports):
    return [ports.replace(':', '-')]



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from neutron_lib import exceptions
from oslo_utils import uuidutils

from vmware_nsx.services.lbaas.octavia import octavia_listener


class DummyOctaviaResource(object):
    create_called = False
    update_called = False
    delete_called = False
    delete_cascade_called = False
    to_raise = False

    def create(self, ctx, lb_obj, completor_func, **args):
        self.create_called = True
        if self.to_raise:
            raise exceptions.InvalidInput(error_message='test')
        else:
            completor_func(success=True)

    def update(self, ctx, old_lb_obj, new_lb_obj, completor_func, **args):
        self.update_called = True
        if self.to_raise:
            raise exceptions.InvalidInput(error_message='test')
        else:
            completor_func(success=True)

    def delete(self, ctx, lb_obj, completor_func, **args):
        self.delete_called = True
        if self.to_raise:
            raise exceptions.InvalidInput(error_message='test')
        else:
            completor_func(success=True)

    def delete_cascade(self, ctx, lb_obj, completor_func, **args):
        self.delete_cascade_called = True
        if self.to_raise:
            raise exceptions.InvalidInput(error_message='test')
        else:
            completor_func(success=True)


class TestNsxOctaviaListener(testtools.TestCase):
    """Test the NSX Octavia listener"""
    def setUp(self):
        super(TestNsxOctaviaListener, self).setUp()
        self.dummyResource = DummyOctaviaResource()
        self.clientMock = mock.Mock()
        self.clientMock.cast = mock.Mock()

        self.endpoint = octavia_listener.NSXOctaviaListenerEndpoint(
            client=self.clientMock,
            loadbalancer=self.dummyResource,
            listener=self.dummyResource,
            pool=self.dummyResource,
            member=self.dummyResource,
            healthmonitor=self.dummyResource,
            l7policy=self.dummyResource,
            l7rule=self.dummyResource)
        self.dummyObj = {'project_id': uuidutils.generate_uuid(),
                         'id': uuidutils.generate_uuid()}
        self.ctx = None
        self.mock_ctx = mock.patch("neutron_lib.context.Context")
        self.mock_ctx.start()

    def tearDown(self):
        self.mock_ctx.stop()
        super(TestNsxOctaviaListener, self).tearDown()

    def test_loadbalancer_create(self):
        self.dummyResource.create_called = False
        self.endpoint.loadbalancer_create(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'loadbalancers': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_loadbalancer_create_failed(self):
        self.dummyResource.create_called = False
        self.dummyResource.to_raise = True
        self.endpoint.loadbalancer_create(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'loadbalancers': [
                {'operating_status': 'ERROR',
                 'provisioning_status': 'ERROR',
                 'id': mock.ANY}]})
        self.dummyResource.to_raise = False

    def test_loadbalancer_delete(self):
        self.dummyResource.delete_called = False
        self.endpoint.loadbalancer_delete(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'loadbalancers': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'DELETED',
                 'id': mock.ANY}]})

    def test_loadbalancer_delete_cascade(self):
        self.dummyResource.delete_called = False
        self.endpoint.loadbalancer_delete(self.ctx, self.dummyObj,
                                          cascade=True)
        self.assertTrue(self.dummyResource.delete_cascade_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={
                'loadbalancers': [{'operating_status': 'ONLINE',
                                   'provisioning_status': 'DELETED',
                 'id': mock.ANY}],
                 'l7policies': [],
                 'pools': [],
                 'listeners': [],
                 'l7rules': [],
                 'members': [],
                 'healthmonitors': []})

    def test_loadbalancer_update(self):
        self.dummyResource.update_called = False
        self.endpoint.loadbalancer_update(self.ctx, self.dummyObj,
                                          self.dummyObj)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'loadbalancers': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_listener_create(self):
        self.dummyResource.create_called = False
        self.endpoint.listener_create(self.ctx, self.dummyObj, None)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'listeners': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_listener_delete(self):
        self.dummyResource.delete_called = False
        self.endpoint.listener_delete(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'listeners': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'DELETED',
                 'id': mock.ANY}]})

    def test_listener_update(self):
        self.dummyResource.update_called = False
        self.endpoint.listener_update(self.ctx, self.dummyObj, self.dummyObj,
                                      None)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'listeners': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_pool_create(self):
        self.dummyResource.create_called = False
        self.endpoint.pool_create(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'pools': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_pool_delete(self):
        self.dummyResource.delete_called = False
        lb_id = uuidutils.generate_uuid()
        listener_id = uuidutils.generate_uuid()
        pool_id = uuidutils.generate_uuid()
        pool_obj = {
            'id': pool_id,
            'pool_id': pool_id,
            'project_id': uuidutils.generate_uuid(),
            'listener_id': listener_id,
            'loadbalancer_id': lb_id,
            'listener': {'protocol': 'HTTP',
                         'id': listener_id,
                         'default_pool_id': pool_id,
                         'loadbalancer': {'id': lb_id}}}
        self.endpoint.pool_delete(self.ctx, pool_obj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'loadbalancers': [{'operating_status': 'ONLINE',
                                       'provisioning_status': 'ACTIVE',
                                       'id': lb_id}],
                    'pools': [{'operating_status': 'ONLINE',
                               'provisioning_status': 'DELETED',
                               'id': pool_id}],
                    'listeners': [{'operating_status': 'ONLINE',
                                   'provisioning_status': 'ACTIVE',
                                   'id': listener_id}],
                    })

    def test_pool_update(self):
        self.dummyResource.update_called = False
        self.endpoint.pool_update(self.ctx, self.dummyObj, self.dummyObj)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'pools': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_member_create(self):
        self.dummyResource.create_called = False
        lb_id = uuidutils.generate_uuid()
        pool_id = uuidutils.generate_uuid()
        listener_id = uuidutils.generate_uuid()
        member_id = uuidutils.generate_uuid()
        member_obj = {
            'id': member_id,
            'protocol_port': 80,
            'name': 'dummy',
            'pool_id': pool_id,
            'project_id': uuidutils.generate_uuid(),
            'pool': {'listener_id': listener_id,
                     'id': pool_id,
                     'loadbalancer_id': lb_id,
                     'listener': {'protocol': 'HTTP',
                                  'id': listener_id,
                                  'default_pool_id': pool_id,
                                  'loadbalancer': {'id': lb_id}}}}
        self.endpoint.member_create(self.ctx, member_obj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'members': [{'operating_status': 'ONLINE',
                                 'provisioning_status': 'ACTIVE',
                                 'id': member_id}],
                    'loadbalancers': [{'operating_status': 'ONLINE',
                                       'provisioning_status': 'ACTIVE',
                                       'id': lb_id}],
                    'pools': [{'operating_status': 'ONLINE',
                               'provisioning_status': 'ACTIVE',
                               'id': pool_id}],
                    'listeners': [{'operating_status': 'ONLINE',
                                   'provisioning_status': 'ACTIVE',
                                   'id': listener_id}],
                    })

    def test_member_delete(self):
        self.dummyResource.delete_called = False
        self.endpoint.member_delete(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'members': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'DELETED',
                 'id': mock.ANY}]})

    def test_member_update(self):
        self.dummyResource.update_called = False
        self.endpoint.member_update(self.ctx, self.dummyObj, self.dummyObj)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'members': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_healthmonitor_create(self):
        self.dummyResource.create_called = False
        self.endpoint.healthmonitor_create(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'healthmonitors': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_healthmonitor_delete(self):
        self.dummyResource.delete_called = False
        self.endpoint.healthmonitor_delete(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'healthmonitors': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'DELETED',
                 'id': mock.ANY}]})

    def test_healthmonitor_update(self):
        self.dummyResource.update_called = False
        self.endpoint.healthmonitor_update(self.ctx, self.dummyObj,
                                           self.dummyObj)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'healthmonitors': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_l7policy_create(self):
        self.dummyResource.create_called = False
        self.endpoint.l7policy_create(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'l7policies': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_l7policy_delete(self):
        self.dummyResource.delete_called = False
        self.endpoint.l7policy_delete(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'l7policies': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'DELETED',
                 'id': mock.ANY}]})

    def test_l7policy_update(self):
        self.dummyResource.update_called = False
        self.endpoint.l7policy_update(self.ctx, self.dummyObj, self.dummyObj)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'l7policies': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_l7rule_create(self):
        self.dummyResource.create_called = False
        self.endpoint.l7rule_create(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.create_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'l7rules': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})

    def test_l7rule_delete(self):
        self.dummyResource.delete_called = False
        self.endpoint.l7rule_delete(self.ctx, self.dummyObj)
        self.assertTrue(self.dummyResource.delete_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'l7rules': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'DELETED',
                 'id': mock.ANY}]})

    def test_l7rule_update(self):
        self.dummyResource.update_called = False
        self.endpoint.l7rule_update(self.ctx, self.dummyObj, self.dummyObj)
        self.assertTrue(self.dummyResource.update_called)
        self.clientMock.cast.assert_called_once_with(
            {}, 'update_loadbalancer_status',
            status={'l7rules': [
                {'operating_status': 'ONLINE',
                 'provisioning_status': 'ACTIVE',
                 'id': mock.ANY}]})



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.tests import base
from neutron_lib import exceptions as n_exc

from vmware_nsx.plugins.common.housekeeper import base_job
from vmware_nsx.plugins.common.housekeeper import housekeeper


class TestJob1(base_job.BaseJob):
    def __init__(self, global_readonly, readonly_jobs):
        super(TestJob1, self).__init__(global_readonly, readonly_jobs)

    def get_name(self):
        return 'test_job1'

    def get_project_plugin(self, plugin):
        return 'Dummy'

    def get_description(self):
        return 'test'

    def run(self, context, readonly=False):
        pass


class TestJob2(TestJob1):
    def get_name(self):
        return 'test_job2'


class TestHousekeeper(base.BaseTestCase):

    def setUp(self):
        self.jobs = ['test_job1', 'test_job2']
        self.readonly_jobs = ['test_job1']
        self.readonly = False
        self.housekeeper = housekeeper.NsxHousekeeper(
            hk_ns='stevedore.test.extension',
            hk_jobs=self.jobs,
            hk_readonly=self.readonly,
            hk_readonly_jobs=self.readonly_jobs)

        self.job1 = TestJob1(self.readonly, self.readonly_jobs)
        self.job2 = TestJob2(self.readonly, self.readonly_jobs)
        self.housekeeper.jobs = {'test_job1': self.job1,
                                 'test_job2': self.job2}
        self.context = mock.Mock()
        self.context.session = mock.Mock()

        super(TestHousekeeper, self).setUp()

    def test_run_job_readonly(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            self.housekeeper.run(self.context, 'test_job1', readonly=True)
            run1.assert_called_with(mock.ANY, readonly=True)

            self.housekeeper.run(self.context, 'test_job2', readonly=True)
            run2.assert_called_with(mock.ANY, readonly=True)

    def test_run_job_readwrite(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            # job1 is configured as a readonly job so this should fail
            self.assertRaises(
                n_exc.ObjectNotFound,
                self.housekeeper.run, self.context, 'test_job1',
                readonly=False)
            self.assertFalse(run1.called)

            # job2 should run
            self.housekeeper.run(self.context, 'test_job2', readonly=False)
            run2.assert_called_with(mock.ANY, readonly=False)

    def test_run_all_readonly(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            self.housekeeper.run(self.context, 'all', readonly=True)
            run1.assert_called_with(mock.ANY, readonly=True)
            run2.assert_called_with(mock.ANY, readonly=True)

    def test_run_all_readwrite(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            self.housekeeper.run(self.context, 'all', readonly=False)
            # job1 is configured as a readonly job so it was not called
            self.assertFalse(run1.called)
            # job2 should run
            run2.assert_called_with(mock.ANY, readonly=False)


class TestHousekeeperReadOnly(TestHousekeeper):

    def setUp(self):
        super(TestHousekeeperReadOnly, self).setUp()
        self.housekeeper.global_readonly = True

    def test_run_job_readonly(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            self.housekeeper.run(self.context, 'test_job1', readonly=True)
            run1.assert_called_with(mock.ANY, readonly=True)

            self.housekeeper.run(self.context, 'test_job2', readonly=True)
            run2.assert_called_with(mock.ANY, readonly=True)

    def test_run_job_readwrite(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:

            # job1 is configured as a readonly job so this should fail
            self.assertRaises(
                n_exc.ObjectNotFound,
                self.housekeeper.run, self.context, 'test_job1',
                readonly=False)
            self.assertFalse(run1.called)

            # global readonly flag so job2 should also fail
            self.assertRaises(
                n_exc.ObjectNotFound,
                self.housekeeper.run, self.context, 'test_job2',
                readonly=False)
            self.assertFalse(run2.called)

    def test_run_all_readonly(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            self.housekeeper.run(self.context, 'all', readonly=True)
            run1.assert_called_with(mock.ANY, readonly=True)
            run2.assert_called_with(mock.ANY, readonly=True)

    def test_run_all_readwrite(self):
        with mock.patch.object(self.job1, 'run') as run1,\
            mock.patch.object(self.job2, 'run') as run2:
            # global readonly flag so 'all' should fail
            self.assertRaises(
                n_exc.ObjectNotFound,
                self.housekeeper.run, self.context, 'all',
                readonly=False)
            self.assertFalse(run1.called)
            self.assertFalse(run2.called)



# Copyright 2016 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock

from neutron_lib import context
from neutron_lib import exceptions
from neutron_lib.objects import registry as obj_reg
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron.services.qos import qos_plugin
from neutron.tests.unit.services.qos import base

from vmware_nsx.common import utils
from vmware_nsx.plugins.nsx_v3 import utils as v3_utils
from vmware_nsx.services.qos.nsx_v3 import driver as qos_driver
from vmware_nsx.services.qos.nsx_v3 import pol_utils as qos_utils
from vmware_nsx.tests.unit.nsx_p import test_plugin
from vmware_nsxlib.v3.policy import core_defs as policy_defs

PLUGIN_NAME = 'vmware_nsx.plugins.nsx_p.plugin.NsxPolicyPlugin'
QoSPolicy = obj_reg.load_class('QosPolicy')
QosBandwidthLimitRule = obj_reg.load_class('QosBandwidthLimitRule')
QosDscpMarkingRule = obj_reg.load_class('QosDscpMarkingRule')
QosMinimumBandwidthRule = obj_reg.load_class('QosMinimumBandwidthRule')


class TestQosNsxPNotification(base.BaseQosTestCase,
                              test_plugin.NsxPPluginTestCaseMixin):

    def setUp(self):
        # Reset the drive to re-create it
        qos_driver.DRIVER = None
        super(TestQosNsxPNotification, self).setUp()
        self.setup_coreplugin(PLUGIN_NAME)

        self.qos_plugin = qos_plugin.QoSPlugin()
        self.ctxt = context.Context('fake_user', 'fake_tenant')
        mock.patch.object(self.ctxt.session, 'refresh').start()
        mock.patch.object(self.ctxt.session, 'expunge').start()
        policy_id = uuidutils.generate_uuid()
        self.project_id = uuidutils.generate_uuid()
        self.policy_data = {
            'policy': {'id': policy_id,
                       'project_id': self.project_id,
                       'name': 'test-policy',
                       'description': 'Test policy description',
                       'shared': True}}
        self.rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': 2000,
                                     'max_burst_kbps': 150}}
        self.ingress_rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': 3000,
                                     'max_burst_kbps': 350,
                                     'direction': 'ingress'}}
        self.dscp_rule_data = {
            'dscp_marking_rule': {'id': uuidutils.generate_uuid(),
                                  'dscp_mark': 22}}

        self.policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])

        # egress BW limit rule
        self.rule = QosBandwidthLimitRule(
            self.ctxt, **self.rule_data['bandwidth_limit_rule'])
        # ingress bw limit rule
        self.ingress_rule = QosBandwidthLimitRule(
            self.ctxt, **self.ingress_rule_data['bandwidth_limit_rule'])
        self.dscp_rule = QosDscpMarkingRule(
            self.ctxt, **self.dscp_rule_data['dscp_marking_rule'])

        self.fake_profile = {'id': policy_id}

        mock.patch('neutron.objects.db.api.create_object').start()
        mock.patch('neutron.objects.db.api.update_object').start()
        mock.patch('neutron.objects.db.api.delete_object').start()

        self.peak_bw_multiplier = cfg.CONF.NSX.qos_peak_bw_multiplier

        self.nsxlib = v3_utils.get_nsxlib_wrapper()

    def _get_expected_tags(self):
        policy_dict = {'id': self.policy.id, 'tenant_id': self.project_id}
        return self.nsxlib.build_v3_tags_payload(
            policy_dict, resource_type='os-neutron-qos-id',
            project_name=self.ctxt.tenant_name)

    @mock.patch.object(QoSPolicy, 'create_rbac_policy')
    def test_policy_create_profile(self, *mocks):
        # test the profile creation when a QoS policy is created
        with mock.patch('vmware_nsxlib.v3.policy.core_resources.'
                        'NsxQosProfileApi.create_or_overwrite',
                        return_value=self.fake_profile) as create_profile,\
            mock.patch.object(QoSPolicy, 'get_object',
                              return_value=self.policy),\
            mock.patch.object(QoSPolicy, 'create'):
            self.qos_plugin.create_policy(self.ctxt, self.policy_data)
            exp_name = utils.get_name_and_uuid(self.policy.name,
                                               self.policy.id)

            create_profile.assert_called_once_with(
                exp_name,
                profile_id=self.policy.id,
                description=self.policy_data["policy"]["description"],
                dscp=None,
                shaper_configurations=[],
                tags=self._get_expected_tags())

    @mock.patch.object(QoSPolicy, '_reload_rules')
    def test_bw_rule_create_profile(self, *mocks):
        # test the profile update when an egress QoS BW rule is created
        _policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [self.rule])
        with mock.patch.object(QoSPolicy, 'get_object', return_value=_policy),\
            mock.patch('vmware_nsxlib.v3.policy.core_resources.'
                       'NsxQosProfileApi.'
                       'create_or_overwrite') as create_profile,\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=self.rule_data):
            self.qos_plugin.update_policy_bandwidth_limit_rule(
                self.ctxt, self.rule.id, _policy.id, self.rule_data)

            # validate the data on the profile
            rule_dict = self.rule_data['bandwidth_limit_rule']
            expected_bw = int(round(float(
                rule_dict['max_kbps']) / 1024))
            expected_burst = rule_dict['max_burst_kbps'] * 128
            expected_peak = int(expected_bw * self.peak_bw_multiplier)
            exp_name = utils.get_name_and_uuid(self.policy.name,
                                               self.policy.id)
            # egress neutron rule -> ingress nsx args
            shaper_type = policy_defs.QoSRateLimiter.INGRESS_RATE_LIMITER_TYPE
            expected_shaper = policy_defs.QoSRateLimiter(
                resource_type=shaper_type,
                enabled=True,
                burst_size=expected_burst,
                peak_bandwidth=expected_peak,
                average_bandwidth=expected_bw)
            create_profile.assert_called_once_with(
                exp_name,
                profile_id=self.policy.id,
                description=self.policy_data["policy"]["description"],
                dscp=None,
                shaper_configurations=[mock.ANY],
                tags=self._get_expected_tags())
            # Compare the shaper
            actual_shaper = create_profile.call_args[1][
                'shaper_configurations'][0]
            self.assertEqual(expected_shaper.get_obj_dict(),
                             actual_shaper.get_obj_dict())

    @mock.patch.object(QoSPolicy, '_reload_rules')
    def test_ingress_bw_rule_create_profile(self, *mocks):
        # test the profile update when a ingress QoS BW rule is created
        _policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [self.ingress_rule])
        with mock.patch.object(QoSPolicy, 'get_object', return_value=_policy),\
            mock.patch('vmware_nsxlib.v3.policy.core_resources.'
                       'NsxQosProfileApi.'
                       'create_or_overwrite') as create_profile,\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=self.ingress_rule_data):
            self.qos_plugin.update_policy_bandwidth_limit_rule(
                self.ctxt, self.ingress_rule.id, _policy.id,
                self.ingress_rule_data)

            # validate the data on the profile
            rule_dict = self.ingress_rule_data['bandwidth_limit_rule']
            expected_bw = int(round(float(
                rule_dict['max_kbps']) / 1024))
            expected_burst = rule_dict['max_burst_kbps'] * 128
            expected_peak = int(expected_bw * self.peak_bw_multiplier)
            exp_name = utils.get_name_and_uuid(self.policy.name,
                                               self.policy.id)
            # ingress neutron rule -> egress nsx args
            shaper_type = policy_defs.QoSRateLimiter.EGRESS_RATE_LIMITER_TYPE
            expected_shaper = policy_defs.QoSRateLimiter(
                resource_type=shaper_type,
                enabled=True,
                burst_size=expected_burst,
                peak_bandwidth=expected_peak,
                average_bandwidth=expected_bw)
            create_profile.assert_called_once_with(
                exp_name,
                profile_id=self.policy.id,
                description=self.policy_data["policy"]["description"],
                dscp=None,
                shaper_configurations=[mock.ANY],
                tags=self._get_expected_tags())
            # Compare the shaper
            actual_shaper = create_profile.call_args[1][
                'shaper_configurations'][0]
            self.assertEqual(expected_shaper.get_obj_dict(),
                             actual_shaper.get_obj_dict())

    @mock.patch.object(QoSPolicy, '_reload_rules')
    def test_bw_rule_create_profile_minimal_val(self, *mocks):
        # test driver precommit with an invalid limit value
        bad_limit = qos_utils.MAX_KBPS_MIN_VALUE - 1
        rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': bad_limit,
                                     'max_burst_kbps': 150}}

        rule = QosBandwidthLimitRule(
            self.ctxt, **rule_data['bandwidth_limit_rule'])

        _policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [rule])
        with mock.patch.object(QoSPolicy, 'get_object',
                               return_value=_policy),\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=rule_data):
            self.assertRaises(
                exceptions.DriverCallError,
                self.qos_plugin.update_policy_bandwidth_limit_rule,
                self.ctxt, rule.id, _policy.id, rule_data)

    @mock.patch.object(QoSPolicy, '_reload_rules')
    def test_bw_rule_create_profile_maximal_val(self, *mocks):
        # test driver precommit with an invalid burst value
        bad_burst = qos_utils.MAX_BURST_MAX_VALUE + 1
        rule_data = {
            'bandwidth_limit_rule': {'id': uuidutils.generate_uuid(),
                                     'max_kbps': 1025,
                                     'max_burst_kbps': bad_burst}}

        rule = QosBandwidthLimitRule(
            self.ctxt, **rule_data['bandwidth_limit_rule'])

        _policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [rule])
        with mock.patch.object(QoSPolicy, 'get_object',
                               return_value=_policy),\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=rule_data):
                self.assertRaises(
                    exceptions.DriverCallError,
                    self.qos_plugin.update_policy_bandwidth_limit_rule,
                    self.ctxt, rule.id, _policy.id, rule_data)

    @mock.patch.object(QoSPolicy, '_reload_rules')
    def test_dscp_rule_create_profile(self, *mocks):
        # test the profile update when a QoS DSCP rule is created
        _policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        # add a rule to the policy
        setattr(_policy, "rules", [self.dscp_rule])
        with mock.patch.object(QoSPolicy, 'get_object', return_value=_policy),\
            mock.patch('vmware_nsxlib.v3.policy.core_resources.'
                       'NsxQosProfileApi.'
                       'create_or_overwrite') as create_profile,\
            mock.patch('neutron.objects.db.api.update_object',
                       return_value=self.dscp_rule_data):
            self.qos_plugin.update_policy_dscp_marking_rule(
                self.ctxt, self.dscp_rule.id,
                _policy.id, self.dscp_rule_data)

            # validate the data on the profile
            rule_dict = self.dscp_rule_data['dscp_marking_rule']
            dscp_mark = rule_dict['dscp_mark']

            exp_name = utils.get_name_and_uuid(self.policy.name,
                                               self.policy.id)
            expected_dscp = policy_defs.QoSDscp(
                mode=policy_defs.QoSDscp.QOS_DSCP_UNTRUSTED,
                priority=dscp_mark)
            create_profile.assert_called_once_with(
                exp_name,
                profile_id=self.policy.id,
                description=self.policy_data["policy"]["description"],
                dscp=mock.ANY,
                shaper_configurations=[],
                tags=self._get_expected_tags())
            # Compare the dscp obj
            actual_dscp = create_profile.call_args[1]['dscp']
            self.assertEqual(expected_dscp.get_obj_dict(),
                             actual_dscp.get_obj_dict())

    @mock.patch.object(QoSPolicy, '_reload_rules')
    def test_minimum_bw_rule_create_profile(self, *mocks):
        # Minimum BW rules are not supported
        policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        min_bw_rule_data = {
            'minimum_bandwidth_rule': {'id': uuidutils.generate_uuid(),
                                       'min_kbps': 10,
                                       'direction': 'egress'}}
        min_bw_rule = QosMinimumBandwidthRule(
            self.ctxt, **min_bw_rule_data['minimum_bandwidth_rule'])
        # add a rule to the policy
        setattr(policy, "rules", [min_bw_rule])
        with mock.patch.object(
                QoSPolicy, 'get_object', return_value=policy),\
            mock.patch('neutron.objects.db.api.'
                       'update_object', return_value=self.dscp_rule_data):
            self.assertRaises(
                exceptions.DriverCallError,
                self.qos_plugin.update_policy_minimum_bandwidth_rule,
                self.ctxt, min_bw_rule.id,
                policy.id, min_bw_rule_data)

    def test_rule_delete_profile(self):
        # test the profile update when a QoS rule is deleted
        _policy = QoSPolicy(
            self.ctxt, **self.policy_data['policy'])
        # The mock will return the policy without the rule,
        # as if it was deleted
        with mock.patch.object(QoSPolicy, 'get_object', return_value=_policy),\
            mock.patch('vmware_nsxlib.v3.policy.core_resources.'
                       'NsxQosProfileApi.'
                       'create_or_overwrite') as set_profile:
            setattr(_policy, "rules", [self.rule])
            self.qos_plugin.delete_policy_bandwidth_limit_rule(
                self.ctxt, self.rule.id, self.policy.id)
            # validate the data on the profile
            exp_name = utils.get_name_and_uuid(self.policy.name,
                                               self.policy.id)

            set_profile.assert_called_once_with(
                exp_name,
                profile_id=self.policy.id,
                description=self.policy_data["policy"]["description"],
                dscp=None,
                shaper_configurations=[],
                tags=self._get_expected_tags())

    @mock.patch('neutron.objects.db.api.get_object', return_value=None)
    def test_policy_delete_profile(self, *mocks):
        # test the profile deletion when a QoS policy is deleted
        with mock.patch('vmware_nsxlib.v3.policy.core_resources.'
                        'NsxQosProfileApi.delete') as delete_profile:
            self.qos_plugin.delete_policy(self.ctxt, self.policy.id)
            delete_profile.assert_called_once_with(self.policy.id)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeL7RuleManagerFromDict(base_mgr.NsxpLoadbalancerBaseManager):
    def _update_l7rule_change(self, rule, completor, delete=False):
        vs_client = self.core_plugin.nsxpolicy.load_balancer.virtual_server
        policy = rule['policy']
        policy_name = utils.get_name_and_uuid(policy['name'] or 'policy',
                                              policy['id'])
        if delete:
            lb_utils.remove_rule_from_policy(rule)
        else:
            lb_utils.update_rule_in_policy(rule)
        rule_body = lb_utils.convert_l7policy_to_lb_rule(
            self.core_plugin.nsxpolicy, rule['policy'])
        try:
            vs_client.update_lb_rule(policy['listener_id'],
                                     policy_name,
                                     position=policy.get('position', 0) - 1,
                                     **rule_body)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to update L7policy %(policy)s: '
                          '%(err)s', {'policy': policy['id'], 'err': e})

        completor(success=True)

    def create(self, context, rule, completor):
        self._update_l7rule_change(rule, completor)

    def update(self, context, old_rule, new_rule, completor):
        self._update_l7rule_change(new_rule, completor)

    def delete(self, context, rule, completor):
        self._update_l7rule_change(rule, completor, delete=True)

    def delete_cascade(self, context, rule, completor):
        # No action should be taken on rules delete cascade
        pass



# Copyright 2019 VMware, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_lib import constants
from neutron_lib import context as n_context

from vmware_nsx.common import exceptions as nsx_exc
from vmware_nsx.services.vpnaas.common_v3 import ipsec_driver as common_driver
from vmware_nsx.services.vpnaas.common_v3 import ipsec_utils
from vmware_nsx.services.vpnaas.nsxp import ipsec_validator
from vmware_nsxlib.v3 import exceptions as nsx_lib_exc
from vmware_nsxlib.v3 import nsx_constants as consts
from vmware_nsxlib.v3.policy import constants as policy_constants

LOG = logging.getLogger(__name__)
IPSEC = 'ipsec'


class NSXpIPsecVpnDriver(common_driver.NSXcommonIPsecVpnDriver):

    def __init__(self, service_plugin):
        validator = ipsec_validator.IPsecNsxPValidator(service_plugin)
        super(NSXpIPsecVpnDriver, self).__init__(service_plugin, validator)

        self._nsxpolicy = self._core_plugin.nsxpolicy
        self._nsx_vpn = self._nsxpolicy.ipsec_vpn

    def _get_service_local_cidr_group(self, context, vpnservice, cidrs):
        """Create/Override the group for the local cidrs of a vpnservice
        used for the edge firewall rules allowing the vpn traffic.
        Return the group id, which is the same as the service id.
        """
        group_id = vpnservice['id']
        expr = self._nsxpolicy.group.build_ip_address_expression(cidrs)
        tags = self._nsxpolicy.build_v3_tags_payload(
            vpnservice,
            resource_type='os-vpn-service-id',
            project_name=context.tenant_name)
        self._nsxpolicy.group.create_or_overwrite_with_conditions(
            "Local group for VPN service %s" % vpnservice['id'],
            policy_constants.DEFAULT_DOMAIN, group_id=group_id,
            conditions=[expr], tags=tags)
        return group_id

    def _delete_service_local_cidr_group(self, vpnservice):
        try:
            self._nsxpolicy.group.delete(
                policy_constants.DEFAULT_DOMAIN, group_id=vpnservice['id'])
        except nsx_lib_exc.ResourceNotFound:
            # If there is no FWaaS on the router it may not have been created
            LOG.debug("Cannot delete local CIDR group for vpnservice %s as "
                      "it was not found", vpnservice['id'])

    def _get_connection_local_cidr_group_name(self, connection):
        return 'local_%s' % connection['id']

    def _get_connection_local_cidr_group(self, context, connection, cidrs):
        """Create/Override the group for the local cidrs of a connection
        used for the edge firewall rules allowing the vpn traffic.
        Return the group id, which is the same as the connection id.
        """
        group_id = self._get_connection_local_cidr_group_name(connection)
        expr = self._nsxpolicy.group.build_ip_address_expression(cidrs)
        tags = self._nsxpolicy.build_v3_tags_payload(
            connection,
            resource_type='os-vpn-connection-id',
            project_name=context.tenant_name)
        self._nsxpolicy.group.create_or_overwrite_with_conditions(
            "Local group for VPN connection %s" % connection['id'],
            policy_constants.DEFAULT_DOMAIN, group_id=group_id,
            conditions=[expr], tags=tags)
        return group_id

    def _delete_connection_local_cidr_group(self, connection):
        try:
            group_id = self._get_connection_local_cidr_group_name(connection)
            self._nsxpolicy.group.delete(
                policy_constants.DEFAULT_DOMAIN, group_id=group_id)
        except nsx_lib_exc.ResourceNotFound:
            # If there is no FWaaS on the router it may not have been created
            LOG.debug("Cannot delete local CIDR group for connection %s as "
                      "it was not found", connection['id'])

    def _get_connection_peer_cidr_group_name(self, connection):
        return 'peer_%s' % connection['id']

    def _get_peer_cidr_group(self, context, conn):
        """Create/Override the group for the peer cidrs of a connection
        used for the edge firewall rules allowing the vpn traffic.
        Return the group id, which is the same as the connection id.
        """
        group_ips = self.validator._get_peer_cidrs(context, conn)
        group_id = self._get_connection_peer_cidr_group_name(conn)
        expr = self._nsxpolicy.group.build_ip_address_expression(group_ips)
        tags = self._nsxpolicy.build_v3_tags_payload(
            conn,
            resource_type='os-vpn-connection-id',
            project_name=context.tenant_name)
        self._nsxpolicy.group.create_or_overwrite_with_conditions(
            "Peer group for VPN connection %s" % conn['id'],
            policy_constants.DEFAULT_DOMAIN, group_id=group_id,
            conditions=[expr], tags=tags)
        return group_id

    def _delete_peer_cidr_group(self, conn):
        try:
            group_id = self._get_connection_peer_cidr_group_name(conn)
            self._nsxpolicy.group.delete(
                policy_constants.DEFAULT_DOMAIN, group_id=group_id)
        except nsx_lib_exc.ResourceNotFound:
            # If there is no FWaaS on the router it may not have been created
            LOG.debug("Cannot delete peer CIDR group for connection %s as "
                      "it was not found", conn['id'])

    def _generate_ipsecvpn_firewall_rules(self, plugin_type, context,
                                          router_id=None):
        """Return the firewall rules needed to allow vpn traffic"""
        fw_rules = []
        # get all the active services of this router
        filters = {'router_id': [router_id],
                   'status': [constants.ACTIVE]}
        services = self.vpn_plugin.get_vpnservices(
            context.elevated(), filters=filters)
        if not services:
            return fw_rules
        for srv in services:
            subnet_id = None
            if srv['subnet_id']:
                subnet_id = srv['subnet_id']
                subnet = self.l3_plugin.get_subnet(
                    context.elevated(), subnet_id)
                local_cidrs = [subnet['cidr']]
                local_group = self._get_service_local_cidr_group(
                    context, srv, local_cidrs)
            # get all the non-errored connections of this service
            filters = {'vpnservice_id': [srv['id']],
                       'status': [constants.ACTIVE, constants.DOWN]}
            connections = self.vpn_plugin.get_ipsec_site_connections(
                context.elevated(), filters=filters)
            for conn in connections:
                if not subnet_id:
                    # Get local endpoint from group
                    local_cidrs = self.validator._get_local_cidrs(
                        context.elevated(), conn)
                    local_group = self._get_connection_local_cidr_group(
                        context, conn, local_cidrs)
                peer_group = self._get_peer_cidr_group(
                    context.elevated(), conn)
                fw_rules.append(self._nsxpolicy.gateway_policy.build_entry(
                    'VPN connection ' + conn['id'],
                    policy_constants.DEFAULT_DOMAIN, router_id,
                    action=consts.FW_ACTION_ALLOW,
                    dest_groups=[peer_group],
                    source_groups=[local_group],
                    scope=[self._nsxpolicy.tier1.get_path(router_id)],
                    direction=consts.IN_OUT))

        return fw_rules

    def _update_firewall_rules(self, context, vpnservice, conn, delete=False):
        LOG.debug("Updating vpn firewall rules for router %s",
                  vpnservice['router_id'])
        self._core_plugin.update_router_firewall(
            context, vpnservice['router_id'])

        # if it is during delete - try to delete the group of this connection
        if delete:
            self._delete_peer_cidr_group(conn)
            self._delete_connection_local_cidr_group(conn)

    def update_router_advertisement(self, context, router_id):
        """Advertise the local subnets of all the services on the router"""

        # Do nothing in case of a router with no GW or no-snat router
        # (as it is already advertised)
        rtr = self.l3_plugin.get_router(context, router_id)
        if (not rtr.get('external_gateway_info') or
            not rtr['external_gateway_info'].get('enable_snat', True)):
            return

        LOG.debug("Updating router advertisement rules for router %s",
                  router_id)
        rules = []

        # get all the active services of this router
        filters = {'router_id': [router_id],
                   'status': [constants.ACTIVE]}
        services = self.vpn_plugin.get_vpnservices(
            context.elevated(), filters=filters)
        rule_name_pref = 'VPN advertisement service'
        has_connections = False
        for srv in services:
            # use only services with non-errored connections
            filters = {'vpnservice_id': [srv['id']],
                       'status': [constants.ACTIVE, constants.DOWN]}
            connections = self.vpn_plugin.get_ipsec_site_connections(
                context.elevated(), filters=filters)
            if not connections:
                continue
            has_connections = True
            if srv['subnet_id']:
                subnet = self.l3_plugin.get_subnet(
                    context.elevated(), srv['subnet_id'])
                local_cidrs = [subnet['cidr']]
            else:
                # get all connections local endpoints cidrs
                local_cidrs = []
                for conn in connections:
                    local_cidrs.extend(
                        self.validator._get_local_cidrs(
                            context.elevated(), conn))
            rules.append(self._nsxpolicy.tier1.build_advertisement_rule(
                "%s %s" % (rule_name_pref, srv['id']),
                policy_constants.ADV_RULE_PERMIT,
                policy_constants.ADV_RULE_OPERATOR_GE,
                [policy_constants.ADV_RULE_TIER1_IPSEC_LOCAL_ENDPOINT],
                local_cidrs))

        self._nsxpolicy.tier1.update_advertisement_rules(
            router_id, rules, name_prefix=rule_name_pref)

        # Also update the ipsec endpoints advertisement
        self._nsxpolicy.tier1.update_route_advertisement(
            router_id, ipsec_endpoints=has_connections)

    def _nsx_tags(self, context, object):
        return self._nsxpolicy.build_v3_tags_payload(
            object, resource_type='os-vpn-connection-id',
            project_name=context.tenant_name)

    def _create_ike_profile(self, context, connection):
        """Create an ike profile for a connection
        Creating/overwriting IKE profile based on the openstack ike policy
        upon connection creation.
        There is no driver callback for profiles creation so it has to be
        done on connection creation.
        """
        ike_policy_id = connection['ikepolicy_id']
        ikepolicy = self.vpn_plugin.get_ikepolicy(context, ike_policy_id)
        tags = self._nsxpolicy.build_v3_tags_payload(
            ikepolicy, resource_type='os-vpn-ikepol-id',
            project_name=context.tenant_name)
        try:
            profile_id = self._nsx_vpn.ike_profile.create_or_overwrite(
                ikepolicy['name'] or ikepolicy['id'],
                profile_id=ikepolicy['id'],
                description=ikepolicy['description'],
                encryption_algorithms=[ipsec_utils.ENCRYPTION_ALGORITHM_MAP[
                    ikepolicy['encryption_algorithm']]],
                digest_algorithms=[ipsec_utils.AUTH_ALGORITHM_MAP_P[
                    ikepolicy['auth_algorithm']]],
                ike_version=ipsec_utils.IKE_VERSION_MAP[
                    ikepolicy['ike_version']],
                dh_groups=[ipsec_utils.PFS_MAP_P[ikepolicy['pfs']]],
                sa_life_time=ikepolicy['lifetime']['value'],
                tags=tags)
        except nsx_lib_exc.ManagerError as e:
            msg = _("Failed to create an ike profile: %s") % e
            raise nsx_exc.NsxPluginException(err_msg=msg)
        return profile_id

    def _delete_ike_profile(self, ikeprofile_id):
        try:
            self._nsx_vpn.ike_profile.delete(ikeprofile_id)
        except nsx_lib_exc.ResourceInUse:
            # Still in use by another connection
            LOG.info("IKE profile %s cannot be deleted yet, because "
                     "another connection still uses it", ikeprofile_id)

    def _create_ipsec_profile(self, context, connection):
        """Create a tunnel profile for a connection
        Creating/overwriting tunnel profile based on the openstack ipsec policy
        upon connection creation.
        There is no driver callback for profiles creation so it has to be
        done on connection creation.
        """
        ipsec_policy_id = connection['ipsecpolicy_id']
        ipsecpolicy = self.vpn_plugin.get_ipsecpolicy(
            context, ipsec_policy_id)
        tags = self._nsxpolicy.build_v3_tags_payload(
            ipsecpolicy, resource_type='os-vpn-ipsecpol-id',
            project_name=context.tenant_name)

        try:
            profile_id = self._nsx_vpn.tunnel_profile.create_or_overwrite(
                ipsecpolicy['name'] or ipsecpolicy['id'],
                profile_id=ipsecpolicy['id'],
                description=ipsecpolicy['description'],
                encryption_algorithms=[ipsec_utils.ENCRYPTION_ALGORITHM_MAP[
                    ipsecpolicy['encryption_algorithm']]],
                digest_algorithms=[ipsec_utils.AUTH_ALGORITHM_MAP_P[
                    ipsecpolicy['auth_algorithm']]],
                dh_groups=[ipsec_utils.PFS_MAP_P[ipsecpolicy['pfs']]],
                sa_life_time=ipsecpolicy['lifetime']['value'],
                tags=tags)
        except nsx_lib_exc.ManagerError as e:
            msg = _("Failed to create a tunnel profile: %s") % e
            raise nsx_exc.NsxPluginException(err_msg=msg)
        return profile_id

    def _delete_ipsec_profile(self, ipsecprofile_id):
        try:
            self._nsx_vpn.tunnel_profile.delete(ipsecprofile_id)
        except nsx_lib_exc.ResourceInUse:
            # Still in use by another connection
            LOG.info("Tunnel profile %s cannot be deleted yet, because "
                     "another connection still uses it", ipsecprofile_id)

    def _create_dpd_profile(self, context, connection):
        """Create a DPD profile for a connection
        Creating/overwriting DPD profile based on the openstack ipsec
        connection configuration upon connection creation.
        There is no driver callback for profiles creation so it has to be
        done on connection creation.
        """
        # TODO(asarfaty) consider reusing profiles based on values
        dpd_info = connection['dpd']
        try:
            profile_id = self._nsx_vpn.dpd_profile.create_or_overwrite(
                self._get_dpd_profile_name(connection),
                profile_id=connection['id'],
                description='neutron dpd profile %s' % connection['id'],
                dpd_probe_interval=dpd_info.get('timeout'),
                enabled=True if dpd_info.get('action') == 'hold' else False,
                tags=self._nsx_tags(context, connection))
        except nsx_lib_exc.ManagerError as e:
            msg = _("Failed to create a DPD profile: %s") % e
            raise nsx_exc.NsxPluginException(err_msg=msg)

        return profile_id

    def _delete_dpd_profile(self, dpdprofile_id):
        self._nsx_vpn.dpd_profile.delete(dpdprofile_id)

    def _update_dpd_profile(self, connection):
        dpd_info = connection['dpd']
        self._nsx_vpn.dpd_profile.update(
            connection['id'],
            name=self._get_dpd_profile_name(connection),
            dpd_probe_interval=dpd_info.get('timeout'),
            enabled=True if dpd_info.get('action') == 'hold' else False)

    def _create_local_endpoint(self, context, connection, vpnservice):
        """Creating/overwrite an NSX local endpoint for a logical router

        This endpoint can be reused by other connections, and will be deleted
        when the router vpn service is deleted.
        """
        # use the router GW as the local ip
        router_id = vpnservice['router']['id']
        local_addr = vpnservice['external_v4_ip']

        # Add the neutron router-id to the tags to help search later
        tags = self._nsxpolicy.build_v3_tags_payload(
            {'id': router_id, 'project_id': vpnservice['project_id']},
            resource_type='os-neutron-router-id',
            project_name=context.tenant_name)

        try:
            ep_client = self._nsx_vpn.local_endpoint
            local_endpoint_id = ep_client.create_or_overwrite(
                'Local endpoint for OS VPNaaS on router %s' % router_id,
                router_id,
                router_id,
                endpoint_id=router_id,
                local_address=local_addr,
                tags=tags)
        except nsx_lib_exc.ManagerError as e:
            msg = _("Failed to create a local endpoint: %s") % e
            raise nsx_exc.NsxPluginException(err_msg=msg)

        return local_endpoint_id

    def _delete_local_endpoint(self, vpnservice):
        router_id = vpnservice['router']['id']
        ctx = n_context.get_admin_context()
        port = self._find_vpn_service_port(ctx, router_id)
        if port:
            self._nsx_vpn.local_endpoint.delete(
                router_id, router_id, router_id)
            self.l3_plugin.delete_port(ctx, port['id'], force_delete_vpn=True)

    def _get_session_rules(self, context, connection):
        peer_cidrs = self.validator._get_peer_cidrs(context, connection)
        local_cidrs = self.validator._get_local_cidrs(context, connection)
        rule = self._nsx_vpn.session.build_rule(
            connection['name'] or connection['id'], connection['id'],
            source_cidrs=local_cidrs, destination_cidrs=peer_cidrs)
        return [rule]

    def _create_session(self, context, connection, vpnservice, local_ep_id,
                        ikeprofile_id, ipsecprofile_id, dpdprofile_id,
                        rules, enabled=True):
        try:
            router_id = vpnservice['router_id']
            session_id = self._nsx_vpn.session.create_or_overwrite(
                connection['name'] or connection['id'],
                tier1_id=router_id,
                vpn_service_id=router_id,
                session_id=connection['id'],
                description=connection['description'],
                peer_address=connection['peer_address'],
                peer_id=connection['peer_id'],
                psk=connection['psk'],
                rules=rules,
                dpd_profile_id=dpdprofile_id,
                ike_profile_id=ikeprofile_id,
                tunnel_profile_id=ipsecprofile_id,
                local_endpoint_id=local_ep_id,
                enabled=enabled,
                tags=self._nsx_tags(context, connection))
        except nsx_lib_exc.ManagerError as e:
            msg = _("Failed to create a session: %s") % e
            raise nsx_exc.NsxPluginException(err_msg=msg)

        return session_id

    def _update_session(self, connection, vpnservice, rules=None,
                        enabled=True):
        router_id = vpnservice['router_id']
        args = {'enabled': enabled}
        if rules is not None:
            args['rules'] = rules
        self._nsx_vpn.session.update(
            router_id, router_id, connection['id'],
            name=connection['name'] or connection['id'],
            description=connection['description'],
            peer_address=connection['peer_address'],
            peer_id=connection['peer_id'],
            psk=connection['psk'],
            **args)

    def get_ipsec_site_connection_status(self, context, ipsec_site_conn_id):
        # find out the router-id of this connection
        conn = self.vpn_plugin._get_ipsec_site_connection(
            context, ipsec_site_conn_id)
        vpnservice_id = conn.vpnservice_id
        vpnservice = self.service_plugin._get_vpnservice(
            context, vpnservice_id)
        router_id = vpnservice['router_id']
        # Get the NSX detailed status
        try:
            status_result = self._nsx_vpn.session.get_status(
                router_id, router_id, ipsec_site_conn_id)
            if status_result and 'results' in status_result:
                status = status_result['results'][0].get('runtime_status', '')
                # NSX statuses are UP, DOWN, DEGRADE
                # VPNaaS connection status should be ACTIVE or DOWN
                if status == 'UP':
                    return 'ACTIVE'
                elif status == 'DOWN' or status == 'DEGRADED':
                    return 'DOWN'
        except nsx_lib_exc.ResourceNotFound:
            LOG.debug("Status for VPN session %s was not found",
                      ipsec_site_conn_id)

    def _delete_session(self, vpnservice, session_id):
        router_id = vpnservice['router_id']
        self._nsx_vpn.session.delete(router_id, router_id, session_id)

    def create_ipsec_site_connection(self, context, ipsec_site_conn):
        LOG.debug('Creating ipsec site connection %(conn_info)s.',
                  {"conn_info": ipsec_site_conn})
        # Note(asarfaty) the plugin already calls the validator
        # which also validated the policies and service

        ikeprofile_id = None
        ipsecprofile_id = None
        dpdprofile_id = None
        session_id = None
        vpnservice_id = ipsec_site_conn['vpnservice_id']
        vpnservice = self.service_plugin._get_vpnservice(
            context, vpnservice_id)
        ipsec_id = ipsec_site_conn["id"]

        try:
            # create the ike profile
            ikeprofile_id = self._create_ike_profile(
                context, ipsec_site_conn)
            LOG.debug("Created NSX ike profile %s", ikeprofile_id)

            # create the ipsec profile
            ipsecprofile_id = self._create_ipsec_profile(
                context, ipsec_site_conn)
            LOG.debug("Created NSX ipsec profile %s", ipsecprofile_id)

            # create the dpd profile
            dpdprofile_id = self._create_dpd_profile(
                context, ipsec_site_conn)
            LOG.debug("Created NSX dpd profile %s", dpdprofile_id)

            # create or reuse a local endpoint using the vpn service
            local_ep_id = self._create_local_endpoint(
                context, ipsec_site_conn, vpnservice)

            # Finally: create the session with policy rules
            rules = self._get_session_rules(context, ipsec_site_conn)
            connection_enabled = (vpnservice['admin_state_up'] and
                                  ipsec_site_conn['admin_state_up'])
            self._create_session(
                context, ipsec_site_conn, vpnservice,
                local_ep_id, ikeprofile_id,
                ipsecprofile_id, dpdprofile_id, rules,
                enabled=connection_enabled)

            self._update_status(context, vpnservice_id, ipsec_id,
                                constants.ACTIVE)

        except nsx_exc.NsxPluginException:
            with excutils.save_and_reraise_exception():
                self._update_status(context, vpnservice_id, ipsec_id,
                                    constants.ERROR)
                # delete the NSX objects that were already created
                # Do not delete reused objects: service, local endpoint
                if session_id:
                    self._delete_session(vpnservice, session_id)
                if dpdprofile_id:
                    self._delete_dpd_profile(dpdprofile_id)
                if ipsecprofile_id:
                    self._delete_ipsec_profile(ipsecprofile_id)
                if ikeprofile_id:
                    self._delete_ike_profile(ikeprofile_id)

        # update router firewall rules
        self._update_firewall_rules(context, vpnservice, ipsec_site_conn)

        # update router advertisement rules
        self.update_router_advertisement(context, vpnservice['router_id'])

    def delete_ipsec_site_connection(self, context, ipsec_site_conn):
        LOG.debug('Deleting ipsec site connection %(site)s.',
                  {"site": ipsec_site_conn})

        vpnservice_id = ipsec_site_conn['vpnservice_id']
        vpnservice = self.service_plugin._get_vpnservice(
            context, vpnservice_id)

        self._delete_session(vpnservice, ipsec_site_conn['id'])
        self._delete_dpd_profile(ipsec_site_conn['id'])
        self._delete_ipsec_profile(ipsec_site_conn['ipsecpolicy_id'])
        self._delete_ike_profile(ipsec_site_conn['ikepolicy_id'])

        # update router firewall rules
        self._update_firewall_rules(context, vpnservice, ipsec_site_conn,
                                    delete=True)

        # update router advertisement rules
        self.update_router_advertisement(context, vpnservice['router_id'])

    def update_ipsec_site_connection(self, context, old_ipsec_conn,
                                     ipsec_site_conn):
        LOG.debug('Updating ipsec site connection new %(site)s.',
                  {"site": ipsec_site_conn})
        LOG.debug('Updating ipsec site connection old %(site)s.',
                  {"site": old_ipsec_conn})

        # Note(asarfaty) the plugin already calls the validator
        # which also validated the policies and service
        # Note(asarfaty): the VPN plugin does not allow changing ike/tunnel
        # policy or the service of a connection during update.
        vpnservice_id = old_ipsec_conn['vpnservice_id']
        vpnservice = self.service_plugin._get_vpnservice(
            context, vpnservice_id)

        # check if the dpd configuration changed
        old_dpd = old_ipsec_conn['dpd']
        new_dpd = ipsec_site_conn['dpd']
        if (old_dpd['action'] != new_dpd['action'] or
            old_dpd['timeout'] != new_dpd['timeout'] or
            old_ipsec_conn['name'] != ipsec_site_conn['name']):
            self._update_dpd_profile(ipsec_site_conn)

        rules = self._get_session_rules(context, ipsec_site_conn)
        connection_enabled = (vpnservice['admin_state_up'] and
                              ipsec_site_conn['admin_state_up'])

        try:
            self._update_session(ipsec_site_conn, vpnservice, rules,
                                 enabled=connection_enabled)
        except nsx_lib_exc.ManagerError as e:
            self._update_status(context, vpnservice_id,
                                ipsec_site_conn['id'],
                                constants.ERROR)
            msg = _("Failed to update VPN session %(id)s: %(error)s") % {
                    "id": ipsec_site_conn['id'], "error": e}
            raise nsx_exc.NsxPluginException(err_msg=msg)

        if (ipsec_site_conn['peer_cidrs'] != old_ipsec_conn['peer_cidrs'] or
            ipsec_site_conn['peer_ep_group_id'] !=
            old_ipsec_conn['peer_ep_group_id']):
            # Update firewall
            self._update_firewall_rules(context, vpnservice, ipsec_site_conn)

        # No service updates. No need to update router advertisement rules

    def _create_vpn_service(self, context, vpnservice):
        """Create or overwrite tier1 vpn service
        The service is created on the TIER1 router attached to the service
        The NSX can keep only one service per tier1 router so we reuse it
        """
        router_id = vpnservice['router_id']
        tags = self._nsxpolicy.build_v3_tags_payload(
            {'id': router_id, 'project_id': vpnservice['project_id']},
            resource_type='os-neutron-router-id',
            project_name=context.tenant_name)

        self._nsx_vpn.service.create_or_overwrite(
            'Neutron VPN service for T1 router ' + router_id,
            router_id,
            vpn_service_id=router_id,
            enabled=True,
            ike_log_level=ipsec_utils.DEFAULT_LOG_LEVEL,
            tags=tags)

    def _should_delete_nsx_service(self, context, vpnservice):
        # Check that no neutron vpn-service is configured for the same router
        router_id = vpnservice['router_id']
        filters = {'router_id': [router_id]}
        services = self.vpn_plugin.get_vpnservices(
            context.elevated(), filters=filters)
        if not services:
            return True

    def _delete_vpn_service(self, context, vpnservice):
        router_id = vpnservice['router_id']
        try:
            self._nsx_vpn.service.delete(router_id, router_id)
        except Exception as e:
            LOG.error("Failed to delete VPN service %s: %s",
                      router_id, e)

        # check if service router should be deleted
        if not self._core_plugin.service_router_has_services(
                context.elevated(), router_id):
            self._core_plugin.delete_service_router(router_id)

    def create_vpnservice(self, context, new_vpnservice):
        LOG.info('Creating VPN service %(vpn)s', {'vpn': new_vpnservice})
        vpnservice_id = new_vpnservice['id']
        vpnservice = self.service_plugin._get_vpnservice(context,
                                                         vpnservice_id)
        try:
            self.validator.validate_vpnservice(context, vpnservice)
            local_address = self._get_service_local_address(
                context.elevated(), vpnservice)
        except Exception:
            with excutils.save_and_reraise_exception():
                # Rolling back change on the neutron
                self.service_plugin.delete_vpnservice(context, vpnservice_id)

        vpnservice['external_v4_ip'] = local_address
        self.service_plugin.set_external_tunnel_ips(context,
                                                    vpnservice_id,
                                                    v4_ip=local_address)

        # Make sure this tier1 has service router
        router_id = vpnservice['router_id']
        if not self._core_plugin.verify_sr_at_backend(router_id):
            self._core_plugin.create_service_router(context, router_id)

        # create the NSX vpn service
        try:
            self._create_vpn_service(context, vpnservice)
        except nsx_lib_exc.ManagerError as e:
            self._update_status(context, vpnservice_id, None, constants.ERROR)
            msg = _("Failed to create vpn service: %s") % e
            raise nsx_exc.NsxPluginException(err_msg=msg)

        # update neutron vpnservice status to active
        self._update_status(context, vpnservice_id, None, constants.ACTIVE)

    def update_vpnservice(self, context, old_vpnservice, vpnservice):
        # Only handle the case of admin-state-up changes
        if old_vpnservice['admin_state_up'] != vpnservice['admin_state_up']:
            # update all relevant connections
            filters = {'vpnservice_id': [vpnservice['id']]}
            connections = self.vpn_plugin.get_ipsec_site_connections(
                context, filters=filters)
            for conn in connections:
                connection_enabled = (vpnservice['admin_state_up'] and
                                      conn['admin_state_up'])
                self._update_session(conn, vpnservice,
                                     enabled=connection_enabled)

    def delete_vpnservice(self, context, vpnservice):
        if self._should_delete_nsx_service(context, vpnservice):
            self._delete_local_endpoint(vpnservice)
            self._delete_vpn_service(context, vpnservice)
        self._delete_service_local_cidr_group(vpnservice)

    def validate_router_gw_info(self, context, router_id, gw_info):
        """Upon router gw update verify no overlapping subnets to advertise"""
        # check if this router has a vpn service
        admin_con = context.elevated()
        # get all relevant services, except those waiting to be deleted or in
        # ERROR state
        filters = {'router_id': [router_id],
                   'status': [constants.ACTIVE, constants.PENDING_CREATE,
                              constants.INACTIVE, constants.PENDING_UPDATE]}
        services = self.vpn_plugin.get_vpnservices(admin_con, filters=filters)
        if not services:
            # This is a non-vpn router. if snat was disabled, should check
            # there is no overlapping with vpn connections advertised
            if (gw_info and
                not gw_info.get('enable_snat',
                                cfg.CONF.enable_snat_by_default)):
                # get router subnets
                subnets = self._core_plugin._find_router_subnets_cidrs(
                    context, router_id)
                # find all vpn services with connections
                if not self._check_subnets_overlap_with_all_conns(
                    admin_con, subnets):
                    raise common_driver.RouterWithOverlapNoSnat(
                        router_id=router_id)



# Copyright 2018 VMware, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
from oslo_utils import excutils

from vmware_nsx._i18n import _
from vmware_nsx.services.lbaas import base_mgr
from vmware_nsx.services.lbaas import lb_common
from vmware_nsx.services.lbaas import lb_const
from vmware_nsx.services.lbaas.nsx_p.implementation import lb_utils
from vmware_nsxlib.v3 import exceptions as nsxlib_exc
from vmware_nsxlib.v3.policy import core_resources
from vmware_nsxlib.v3.policy import lb_defs
from vmware_nsxlib.v3.policy import utils as p_utils
from vmware_nsxlib.v3 import utils

LOG = logging.getLogger(__name__)


class EdgeListenerManagerFromDict(base_mgr.NsxpLoadbalancerBaseManager):
    def _get_listener_tags(self, context, listener):
        tags = lb_utils.get_tags(self.core_plugin, listener['id'],
                                 lb_const.LB_LISTENER_TYPE,
                                 listener.get('tenant_id'),
                                 context.project_name)
        if listener['loadbalancer'].get('name'):
            tags.append({
                'scope': lb_const.LB_LB_NAME,
                'tag': listener['loadbalancer']['name'][:utils.MAX_TAG_LEN]})
        tags.append({
            'scope': lb_const.LB_LB_TYPE,
            'tag': listener['loadbalancer_id']})
        return tags

    def _upload_certificate(self, listener_id, cert_href, tags,
                            certificate):
        nsxpolicy = self.core_plugin.nsxpolicy
        cert_client = nsxpolicy.certificate
        ssl_client = nsxpolicy.load_balancer.client_ssl_profile

        # check if this certificate was already uploaded
        cert_ids = cert_client.find_cert_with_pem(
            certificate.get('certificate'))
        if cert_ids:
            nsx_cert_id = cert_ids[0]
        else:
            # Create it with a random id as this might not be the first one
            passphrase = certificate.get('passphrase')
            if not passphrase:
                passphrase = core_resources.IGNORE
            nsx_cert_id = cert_client.create_or_overwrite(
                cert_href,
                pem_encoded=certificate.get('certificate'),
                private_key=certificate.get('private_key'),
                passphrase=passphrase,
                tags=tags)

        return {
            'client_ssl_profile_binding': {
                'ssl_profile_path': ssl_client.get_path(
                    self.core_plugin.client_ssl_profile),
                'default_certificate_path': cert_client.get_path(nsx_cert_id)
            }
        }

    def _get_virtual_server_kwargs(self, context, listener, vs_name, tags,
                                   certificate=None):
        # If loadbalancer vip_port already has floating ip, use floating
        # IP as the virtual server VIP address. Else, use the loadbalancer
        # vip_address directly on virtual server.
        filters = {'port_id': [listener['loadbalancer']['vip_port_id']]}
        floating_ips = self.core_plugin.get_floatingips(context,
                                                        filters=filters)
        if floating_ips:
            lb_vip_address = floating_ips[0]['floating_ip_address']
        else:
            lb_vip_address = listener['loadbalancer']['vip_address']
        lb_service = lb_utils.get_lb_nsx_lb_service(
            self.core_plugin.nsxpolicy, listener['loadbalancer_id'])

        kwargs = {'virtual_server_id': listener['id'],
                  'ip_address': lb_vip_address,
                  'ports': [listener['protocol_port']],
                  'application_profile_id': listener['id'],
                  'lb_service_id': lb_service['id'],
                  'description': listener.get('description')}
        if vs_name:
            kwargs['name'] = vs_name
        if tags:
            kwargs['tags'] = tags
        if listener['connection_limit'] != -1:
            kwargs['max_concurrent_connections'] = listener['connection_limit']
        if 'default_pool_id' in listener:
            if listener['default_pool_id']:
                kwargs['pool_id'] = listener['default_pool_id']
            else:
                # Remove the default pool
                kwargs['pool_id'] = ''
                kwargs['lb_persistence_profile_id'] = ''
        if certificate:
            ssl_profile_binding = self._upload_certificate(
                listener['id'], certificate['ref'], tags,
                certificate=certificate)
            if (listener['protocol'] == lb_const.LB_PROTOCOL_TERMINATED_HTTPS
                and ssl_profile_binding):
                kwargs.update(ssl_profile_binding)

        waf_profile, mode = self.core_plugin.get_waf_profile_path_and_mode()
        if (waf_profile and (
            listener['protocol'] == lb_const.LB_PROTOCOL_HTTP or
            listener['protocol'] == lb_const.LB_PROTOCOL_TERMINATED_HTTPS)):
            kwargs['waf_profile_binding'] = lb_defs.WAFProfileBindingDef(
                waf_profile_path=waf_profile,
                operational_mode=mode)

        return kwargs

    def _get_nsxlib_app_profile(self, nsxlib_lb, listener):
        if (listener['protocol'] == lb_const.LB_PROTOCOL_HTTP or
                listener['protocol'] == lb_const.LB_PROTOCOL_TERMINATED_HTTPS):
            app_client = nsxlib_lb.lb_http_profile
        elif (listener['protocol'] == lb_const.LB_PROTOCOL_TCP or
              listener['protocol'] == lb_const.LB_PROTOCOL_HTTPS):
            app_client = nsxlib_lb.lb_fast_tcp_profile
        else:
            msg = (_('Cannot create listener %(listener)s with '
                     'protocol %(protocol)s') %
                   {'listener': listener['id'],
                    'protocol': listener['protocol']})
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        return app_client

    def _validate_default_pool(self, listener, completor):
        def_pool_id = listener.get('default_pool_id')
        if def_pool_id:
            vs_client = self.core_plugin.nsxpolicy.load_balancer.virtual_server
            vs_list = vs_client.list()
            for vs in vs_list:
                if vs.get('id') == listener['id']:
                    continue
                pool_id = p_utils.path_to_id(vs.get('pool_path', ''))
                if pool_id == def_pool_id:
                    completor(success=False)
                    msg = (_('Default pool %(p)s is already used by another '
                             'listener %(l)s') % {'p': def_pool_id,
                                                  'l': vs.get('id')})
                    raise n_exc.BadRequest(resource='lbaas-pool', msg=msg)

            lb_common.validate_session_persistence(
                listener.get('default_pool'), listener, completor)

    def create(self, context, listener, completor, certificate=None):
        nsxlib_lb = self.core_plugin.nsxpolicy.load_balancer
        vs_client = nsxlib_lb.virtual_server
        vs_name = utils.get_name_and_uuid(listener['name'] or 'listener',
                                          listener['id'])
        tags = self._get_listener_tags(context, listener)
        self._validate_default_pool(listener, completor)
        try:
            app_client = self._get_nsxlib_app_profile(nsxlib_lb, listener)
            app_client.create_or_overwrite(
                lb_app_profile_id=listener['id'], name=vs_name, tags=tags)
            kwargs = self._get_virtual_server_kwargs(
                context, listener, vs_name, tags, certificate)
            vs_client.create_or_overwrite(**kwargs)
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = _('Failed to create virtual server at NSX backend')
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        self._update_default_pool(context, listener, completor)

        completor(success=True)

    def _get_pool_tags(self, context, pool, listener_tenant_id):
        return lb_utils.get_tags(self.core_plugin, pool['id'],
                                 lb_const.LB_POOL_TYPE,
                                 pool.get('tenant_id', listener_tenant_id),
                                 context.project_name)

    def _update_default_pool(self, context, listener,
                             completor, old_listener=None):
        if not listener.get('default_pool_id'):
            return
        nsxlib_lb = self.core_plugin.nsxpolicy.load_balancer
        vs_client = nsxlib_lb.virtual_server
        vs_data = vs_client.get(listener['id'])
        pool_id = listener['default_pool_id']
        pool = listener['default_pool']
        old_pool = None
        if old_listener:
            old_pool = old_listener.get('default_pool')
        try:
            switch_type = lb_common.session_persistence_type_changed(
                pool, old_pool)
            (persistence_profile_id,
             post_process_func) = lb_utils.setup_session_persistence(
                self.core_plugin.nsxpolicy,
                pool,
                self._get_pool_tags(context, pool, listener.get('tenant_id')),
                switch_type, listener, vs_data)
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error("Failed to configure session persistence "
                          "profile for listener %s", listener['id'])
        try:
            # Update persistence profile and pool on virtual server
            vs_client.update(
                listener['id'],
                pool_id=pool_id,
                lb_persistence_profile_id=persistence_profile_id)
            LOG.debug("Updated NSX virtual server %(vs_id)s with "
                      "persistence profile %(prof)s",
                      {'vs_id': listener['id'],
                       'prof': persistence_profile_id})
            if post_process_func:
                post_process_func()
        except nsxlib_exc.ManagerError:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error("Failed to attach persistence profile %s to "
                          "virtual server %s",
                          persistence_profile_id, listener['id'])

    def update(self, context, old_listener, new_listener, completor,
               certificate=None):
        nsxlib_lb = self.core_plugin.nsxpolicy.load_balancer
        vs_client = nsxlib_lb.virtual_server
        app_client = self._get_nsxlib_app_profile(nsxlib_lb, old_listener)

        vs_name = None
        self._validate_default_pool(new_listener, completor)

        if new_listener['name'] != old_listener['name']:
            vs_name = utils.get_name_and_uuid(
                new_listener['name'] or 'listener',
                new_listener['id'])
        tags = self._get_listener_tags(context, new_listener)

        try:
            app_profile_id = new_listener['id']
            updated_kwargs = self._get_virtual_server_kwargs(
                context, new_listener, vs_name, tags, certificate)
            vs_client.update(**updated_kwargs)
            if vs_name:
                app_client.update(app_profile_id, name=vs_name,
                                  tags=tags)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                completor(success=False)
                LOG.error('Failed to update listener %(listener)s with '
                          'error %(error)s',
                          {'listener': old_listener['id'], 'error': e})

        # Update default pool and session persistence
        if (old_listener.get('default_pool_id') !=
            new_listener.get('default_pool_id')):
            self._update_default_pool(context, new_listener,
                                      completor, old_listener)
        completor(success=True)

    def delete(self, context, listener, completor):
        nsxlib_lb = self.core_plugin.nsxpolicy.load_balancer
        vs_client = nsxlib_lb.virtual_server
        app_client = self._get_nsxlib_app_profile(nsxlib_lb, listener)

        vs_id = listener['id']
        app_profile_id = listener['id']

        try:
            profile_path = None
            if listener.get('default_pool_id'):
                vs_data = vs_client.get(vs_id)
                profile_path = vs_data.get('lb_persistence_profile_path', '')
            vs_client.delete(vs_id)
            # Also delete the old session persistence profile
            if profile_path:
                lb_utils.delete_persistence_profile(
                    self.core_plugin.nsxpolicy, profile_path)
        except nsxlib_exc.ResourceNotFound:
            LOG.error("virtual server not found on nsx: %(vs)s", {'vs': vs_id})
        except nsxlib_exc.ManagerError:
            completor(success=False)
            msg = (_('Failed to delete virtual server: %(vs)s') %
                   {'vs': vs_id})
            raise n_exc.BadRequest(resource='lbaas-listener', msg=msg)

        try:
            app_client.delete(app_profile_id)
        except nsxlib_exc.ResourceNotFound:
            LOG.error("application profile not found on nsx: %s",
                      app_profile_id)
        except nsxlib_exc.ManagerError as e:
            # This probably means that the application profile is being
            # used by a listener outside of openstack
            LOG.error("Failed to delete application profile %s from the "
                      "NSX: %s", app_profile_id, e)

        # Delete imported NSX certificates if there is any
        cert_client = self.core_plugin.nsxpolicy.certificate
        cert_tags = [{'scope': lb_const.LB_LISTENER_TYPE,
                      'tag': listener['id']}]
        results = self.core_plugin.nsxpolicy.search_by_tags(
            cert_tags, cert_client.entry_def.resource_type())
        for res_obj in results['results']:
            try:
                cert_client.delete(res_obj['id'])
            except nsxlib_exc.ManagerError:
                msg = (_('Failed to delete certificate: %(crt)s for '
                         'listener %(list)s') %
                       {'crt': res_obj['id'], 'list': listener['id']})
                LOG.error(msg)

        completor(success=True)

    def delete_cascade(self, context, listener, completor):
        self.delete(context, listener, completor)


def stats_getter(context, core_plugin, ignore_list=None):
    """Update Octavia statistics for each listener (virtual server)"""
    stat_list = []
    lb_service_client = core_plugin.nsxpolicy.load_balancer.lb_service

    lb_services = lb_service_client.list()
    # Go over all the loadbalancers & services
    for lb_service in lb_services:
        if ignore_list and lb_service['id'] in ignore_list:
            continue

        lb_service_id = lb_service.get('id')
        try:
            # get the NSX statistics for this LB service
            stats_results = lb_service_client.get_statistics(
                lb_service_id).get('results', [])
            if stats_results:
                rsp = stats_results[0]
            else:
                rsp = {}

            # Go over each virtual server in the response
            for vs in rsp.get('virtual_servers', []):
                # look up the virtual server in the DB
                if vs.get('statistics'):
                    vs_stats = vs['statistics']
                    stats = copy.copy(lb_const.LB_EMPTY_STATS)
                    stats['id'] = p_utils.path_to_id(
                        vs['virtual_server_path'])
                    stats['request_errors'] = 0  # currently unsupported
                    for stat in lb_const.LB_STATS_MAP:
                        lb_stat = lb_const.LB_STATS_MAP[stat]
                        stats[stat] += vs_stats[lb_stat]
                    stat_list.append(stats)

        except nsxlib_exc.ManagerError:
            pass

    return stat_list



# Copyright 2016 VMware, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from vmware_nsx.common import utils as nsx_utils
from vmware_nsx.db import db as nsx_db
from vmware_nsx.dvs import dvs
from vmware_nsx.plugins.nsx_v3 import plugin
from vmware_nsx.plugins.nsx_v3 import utils as plugin_utils
from vmware_nsx.shell.admin.plugins.common import constants
from vmware_nsx.shell.admin.plugins.common import formatters
from vmware_nsx.shell.admin.plugins.common import utils as admin_utils
from vmware_nsx.shell.admin.plugins.nsxv3.resources import utils as v3_utils
from vmware_nsx.shell import resources as shell
from vmware_nsxlib.v3 import exceptions as nsx_exc
from vmware_nsxlib.v3 import nsx_constants as nsxlib_consts
from vmware_nsxlib.v3 import resources
from vmware_nsxlib.v3 import security

from neutron.db import allowedaddresspairs_db as addr_pair_db
from neutron.db import db_base_plugin_v2
from neutron.db import l3_db
from neutron.db import portsecurity_db
from neutron_lib.callbacks import registry
from neutron_lib import constants as const
from neutron_lib import context as neutron_context
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory

LOG = logging.getLogger(__name__)


class PortsPlugin(db_base_plugin_v2.NeutronDbPluginV2,
                  portsecurity_db.PortSecurityDbMixin,
                  addr_pair_db.AllowedAddressPairsMixin):
    def __enter__(self):
        directory.add_plugin(plugin_constants.CORE, self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        directory.add_plugin(plugin_constants.CORE, None)


def get_network_nsx_id(session, neutron_id):
    # get the nsx switch id from the DB mapping
    mappings = nsx_db.get_nsx_switch_ids(session, neutron_id)
    if not mappings or len(mappings) == 0:
        LOG.debug("Unable to find NSX mappings for neutron "
                  "network %s.", neutron_id)
        # fallback in case we didn't find the id in the db mapping
        # This should not happen, but added here in case the network was
        # created before this code was added.
        return neutron_id
    else:
        return mappings[0]


@admin_utils.output_header
def list_missing_ports(resource, event, trigger, **kwargs):
    """List neutron ports that are missing the NSX backend port
    And ports with wrong switch profiles or bindings
    """
    admin_cxt = neutron_context.get_admin_context()
    filters = v3_utils.get_plugin_filters(admin_cxt)
    nsxlib = v3_utils.get_connected_nsxlib()
    with v3_utils.NsxV3PluginWrapper() as plugin:
        problems = plugin_utils.get_mismatch_logical_ports(
            admin_cxt, nsxlib, plugin, filters)

    if len(problems) > 0:
        title = ("Found internal ports misconfiguration on the "
                 "NSX manager:")
        LOG.info(formatters.output_formatter(
            title, problems,
            ['neutron_id', 'nsx_id', 'error']))
    else:
        LOG.info("All internal ports verified on the NSX manager")


def get_vm_network_device(vm_mng, vm_moref, mac_address):
    """Return the network device with MAC 'mac_address'.

    This code was inspired by Nova vif.get_network_device
    """
    hardware_devices = vm_mng.get_vm_interfaces_info(vm_moref)
    if hardware_devices.__class__.__name__ == "ArrayOfVirtualDevice":
        hardware_devices = hardware_devices.VirtualDevice
    for device in hardware_devices:
        if hasattr(device, 'macAddress'):
            if device.macAddress == mac_address:
                return device


def migrate_compute_ports_vms(resource, event, trigger, **kwargs):
    """Update the VMs ports on the backend after migrating nsx-v -> nsx-v3

    After using api_replay to migrate the neutron data from NSX-V to NSX-T
    we need to update the VM ports to use OpaqueNetwork instead of
    DistributedVirtualPortgroup
    """
    # Connect to the DVS manager, using the configuration parameters
    try:
        vm_mng = dvs.VMManager()
    except Exception as e:
        LOG.error("Cannot connect to the DVS: Please update the [dvs] "
                  "section in the nsx.ini file: %s", e)
        return

    port_filters = {}
    if kwargs.get('property'):
        properties = admin_utils.parse_multi_keyval_opt(kwargs['property'])
        project = properties.get('project-id')
        if project:
            port_filters['project_id'] = [project]
        net_name = properties.get('net-name', 'VM Network')
        LOG.info("Common network name for migration %s", net_name)
        host_moref = properties.get('host-moref')
        # TODO(garyk): We can explore the option of passing the cluster
        # moref then this will remove the need for the host-moref and the
        # resource pool moref.
        respool_moref = properties.get('respool-moref')
        datastore_moref = properties.get('datastore-moref')
        if not host_moref:
            LOG.error("Unable to migrate with no host")
            return

    # Go over all the ports from the plugin
    admin_cxt = neutron_context.get_admin_context()
    with PortsPlugin() as plugin:
        neutron_ports = plugin.get_ports(admin_cxt, filters=port_filters)

    for port in neutron_ports:
        # skip non compute ports
        if (not port.get('device_owner').startswith(
            const.DEVICE_OWNER_COMPUTE_PREFIX)):
            continue
        device_id = port.get('device_id')

        # get the vm moref & spec from the DVS
        vm_moref = vm_mng.get_vm_moref_obj(device_id)
        vm_spec = vm_mng.get_vm_spec(vm_moref)
        if not vm_spec:
            LOG.error("Failed to get the spec of vm %s", device_id)
            continue

        # Go over the VM interfaces and check if it should be updated
        update_spec = False
        for prop in vm_spec.propSet:
            if (prop.name == 'network' and
                hasattr(prop.val, 'ManagedObjectReference')):
                for net in prop.val.ManagedObjectReference:
                    if (net._type == 'DistributedVirtualPortgroup' or
                        net._type == 'Network'):
                        update_spec = True

        if not update_spec:
            LOG.info("No need to update the spec of vm %s", device_id)
            continue

        device = get_vm_network_device(vm_mng, vm_moref, port['mac_address'])
        if device is None:
            LOG.warning("No device with MAC address %s exists on the VM",
                        port['mac_address'])
            continue

        # Update interface to be common network
        devices = [vm_mng.update_vm_network(device, name=net_name)]
        LOG.info("Update instance %s to common network", device_id)
        vm_mng.update_vm_interface(vm_moref, devices=devices)
        LOG.info("Migrate instance %s to host %s", device_id, host_moref)
        vm_mng.relocate_vm(vm_moref, host_moref=host_moref,
                           datastore_moref=datastore_moref,
                           respool_moref=respool_moref)
        LOG.info("Update instance %s to opaque network", device_id)
        device = get_vm_network_device(vm_mng, vm_moref, port['mac_address'])
        vif_info = {'nsx_id': get_network_nsx_id(admin_cxt.session,
                                                 port['network_id']),
                    'iface_id': port['id']}
        devices = [vm_mng.update_vm_opaque_spec(vif_info, device)]
        vm_mng.update_vm_interface(vm_moref, devices=devices)
        LOG.info("Instance %s successfully migrated!", device_id)


def migrate_exclude_ports(resource, event, trigger, **kwargs):
    _nsx_client = v3_utils.get_nsxv3_client()

    nsxlib = v3_utils.get_connected_nsxlib()
    version = nsxlib.get_version()
    if not nsx_utils.is_nsx_version_2_0_0(version):
        LOG.info("Migration only supported from 2.0 onwards")
        LOG.info("Version is %s", version)
        return
    admin_cxt = neutron_context.get_admin_context()
    plugin = PortsPlugin()
    _port_client = resources.LogicalPort(_nsx_client)
    exclude_list = nsxlib.firewall_section.get_excludelist()
    for member in exclude_list['members']:
        if member['target_type'] == 'LogicalPort':
            port_id = member['target_id']
            # Get port
            try:
                nsx_port = _port_client.get(port_id)
            except nsx_exc.ResourceNotFound:
                LOG.info("Port %s not found", port_id)
                continue
            # Validate its a neutron port
            is_neutron_port = False
            for tag in nsx_port.get('tags', []):
                if tag['scope'] == 'os-neutron-port-id':
                    is_neutron_port = True
                    neutron_port_id = tag['tag']
                    break
            if not is_neutron_port:
                LOG.info("Port %s is not a neutron port", port_id)
                continue
            # Check if this port exists in the DB
            try:
                plugin.get_port(admin_cxt, neutron_port_id)
            except Exception:
                LOG.info("Port %s is not defined in DB", neutron_port_id)
                continue
            # Update tag for the port
            tags_update = [{'scope': security.PORT_SG_SCOPE,
                            'tag': nsxlib_consts.EXCLUDE_PORT}]
            _port_client.update(port_id, None,
                                tags_update=tags_update)
            # Remove port from the exclude list
            nsxlib.firewall_section.remove_member_from_fw_exclude_list(
                port_id, nsxlib_consts.TARGET_TYPE_LOGICAL_PORT)
            LOG.info("Port %s successfully updated", port_id)


def tag_default_ports(resource, event, trigger, **kwargs):
    nsxlib = v3_utils.get_connected_nsxlib()
    admin_cxt = neutron_context.get_admin_context()
    filters = v3_utils.get_plugin_filters(admin_cxt)

    # the plugin creation below will create the NS group and update the default
    # OS section to have the correct applied to group
    with v3_utils.NsxV3PluginWrapper() as _plugin:
        neutron_ports = _plugin.get_ports(admin_cxt, filters=filters)
        for port in neutron_ports:
            neutron_id = port['id']
            # get the network nsx id from the mapping table
            nsx_id = plugin_utils.get_port_nsx_id(admin_cxt.session,
                                                  neutron_id)
            if not nsx_id:
                continue
            device_owner = port['device_owner']
            if (device_owner == l3_db.DEVICE_OWNER_ROUTER_INTF or
                device_owner == const.DEVICE_OWNER_DHCP):
                continue
            ps = _plugin._get_port_security_binding(admin_cxt,
                                                    neutron_id)
            if not ps:
                continue
            try:
                nsx_port = nsxlib.logical_port.get(nsx_id)
            except nsx_exc.ResourceNotFound:
                continue
            tags_update = nsx_port['tags']
            tags_update += [{'scope': security.PORT_SG_SCOPE,
                             'tag': plugin.NSX_V3_DEFAULT_SECTION}]
            nsxlib.logical_port.update(nsx_id, None,
                                       tags_update=tags_update)


registry.subscribe(list_missing_ports,
                   constants.PORTS,
                   shell.Operations.LIST_MISMATCHES.value)

registry.subscribe(migrate_compute_ports_vms,
                   constants.PORTS,
                   shell.Operations.NSX_MIGRATE_V_V3.value)

registry.subscribe(migrate_exclude_ports,
                   constants.PORTS,
                   shell.Operations.NSX_MIGRATE_EXCLUDE_PORTS.value)


registry.subscribe(tag_default_ports,
                   constants.PORTS,
                   shell.Operations.NSX_TAG_DEFAULT.value)
